class Logger(object):
    def __init__(self, logger):
        self.logger = logger

    def debug(self, message):
        try:
            self.logger.debug(message)
        except:
            print("DEBUG: %s" % message)

    def info(self, message):
        try:
            self.logger.info(message)
        except:
            print("INFO: %s" % message)

    def warning(self, message):
        try:
            self.logger.warning(message)
        except:
            print("WARNING: %s" % message)

    def error(self, message):
        try:
            self.logger.error(message)
        except:
            print("ERROR: %s" % message)



from unittest import TestCase
from unittest.mock import patch, MagicMock
from pyworker.reporter import Reporter


class TestReporter(TestCase):

    def setUp(self) -> None:
        return super().setUp()

    def tearDown(self) -> None:
        return super().tearDown()

    #********** __init__ tests **********#

    @patch('pyworker.reporter.newrelic.agent')
    def test_reporter_init_initializes_newrelic(self, newrelic_agent):
        newrelic_app = MagicMock()
        newrelic_agent.register_application.return_value = newrelic_app
        reporter = Reporter(attribute_prefix='test_prefix')

        self.assertEqual(reporter._prefix, 'test_prefix')
        self.assertIsNone(reporter._logger)
        self.assertEqual(reporter._newrelic_app, newrelic_app)
        newrelic_agent.register_application.assert_called_once_with()
        newrelic_agent.initialize.assert_called_once_with()

    #********** .report tests **********#

    @patch('pyworker.reporter.Reporter._report_newrelic')
    @patch('pyworker.reporter.Reporter._format_attributes')
    @patch('pyworker.reporter.Reporter._flatten_attributes')
    def test_reporter_report_formats_attributes_and_reports_to_newrelic(self,
            mock_flatten_attributes, mock_format_attributes, mock_report_newrelic):
        reporter = Reporter()
        mock_flatten_attributes.return_value = {'flattened_attribute': '1'}
        mock_format_attributes.return_value = {'formatted_attribute': '1'}
        reporter.report(test_attribute={'nested_attribute': 1})

        mock_flatten_attributes.assert_called_once_with({'test_attribute': {'nested_attribute': 1}})
        mock_format_attributes.assert_called_once_with({'flattened_attribute': '1'})
        mock_report_newrelic.assert_called_once_with({'formatted_attribute': '1'})

    #********** .report_raw tests **********#

    @patch('pyworker.reporter.Reporter._report_newrelic')
    def test_reporter_report_raw_reports_to_newrelic(self, mock_report_newrelic):
        reporter = Reporter()
        reporter.report_raw(test_attribute=1)

        mock_report_newrelic.assert_called_once_with({'test_attribute': 1})

    #********** .recorder tests **********#

    @patch('pyworker.reporter.newrelic.agent')
    def test_reporter_recorder_returns_newrelic_background_task(self, newrelic_agent):
        reporter = Reporter()

        with reporter.recorder('test_name'):
            newrelic_agent.BackgroundTask.assert_called_once_with(
                application=reporter._newrelic_app,
                name='test_name',
                group='DelayedJob'
            )

    #********** .shutdown tests **********#

    @patch('pyworker.reporter.newrelic.agent')
    def test_reporter_shutdown_calls_newrelic_shutdown_agent(self, newrelic_agent):
        reporter = Reporter()
        reporter.shutdown()

        newrelic_agent.shutdown_agent.assert_called_once_with()

    #********** .record_exception tests **********#

    @patch('pyworker.reporter.newrelic.agent')
    def test_reporter_record_exception_calls_newrelic_notice_error(self, newrelic_agent):
        reporter = Reporter()
        mock_exc_info = ('test_exception', 'test_value', 'test_traceback')
        reporter.record_exception(mock_exc_info)

        newrelic_agent.notice_error.assert_called_once_with(error=mock_exc_info)

    #********** ._flatten_attributes tests **********#

    def test_reporter_flatten_attributes_flattens_nested_dict_attributes(self):
        reporter = Reporter()

        self.assertEqual(
            reporter._flatten_attributes({
                'test_key1': 'test_value1',
                'test_key2': 2,
                'test_key3': 3.0,
                'test_key4': True,
                'test_key5': {'test_key6': [1, 'test_value6']},
                'test_key7': None,
                None: 'test_value8'
            }),
            {
                'test_key1': 'test_value1',
                'test_key2': 2,
                'test_key3': 3.0,
                'test_key4': True,
                'test_key6': [1, 'test_value6'],
                'test_key7': None,
                None: 'test_value8'
            }
        )

    #********** ._format_attributes tests **********#

    def test_reporter_format_attributes_prefixes_and_camel_cases_keys_and_converts_values(self):
        reporter = Reporter(attribute_prefix='prefix.')

        self.assertEqual(
            reporter._format_attributes({
                'test_key1': 'test_value1',
                'test_key2': 2,
                'test_key3': 3.0,
                'test_key4': True,
                'test_key5': [1, 'test_value6'],
                'test_key7': None,
                None: 'test_value8'
            }),
            {
                'prefix.testKey1': 'test_value1',
                'prefix.testKey2': 2,
                'prefix.testKey3': 3.0,
                'prefix.testKey4': True,
                'prefix.testKey5': '[1, "test_value6"]'
            }
        )

    #********** ._to_camel_case tests **********#

    def test_reporter_to_camel_case_converts_to_camel_case(self):
        reporter = Reporter()

        self.assertEqual(
            reporter._to_camel_case('test-key'),
            'testKey'
        )
        self.assertEqual(
            reporter._to_camel_case('test_key'),
            'testKey'
        )
        self.assertEqual(
            reporter._to_camel_case('test key'),
            'testKey'
        )

    #********** ._convert_value tests **********#

    def test_reporter_convert_value_converts_value_to_json_if_not_supported(self):
        reporter = Reporter()

        self.assertEqual(
            reporter._convert_value({'test_key': 'test_value'}),
            '{"test_key": "test_value"}'
        )
        self.assertEqual(
            reporter._convert_value('test_value'),
            'test_value'
        )
        self.assertEqual(
            reporter._convert_value(1),
            1
        )
        self.assertEqual(
            reporter._convert_value(1.0),
            1.0
        )
        self.assertEqual(
            reporter._convert_value(True),
            True
        )
        self.assertEqual(
            reporter._convert_value(False),
            False
        )
        self.assertEqual(
            reporter._convert_value(None),
            'null'
        )

    #********** ._report_newrelic tests **********#

    @patch('pyworker.reporter.newrelic.agent')
    def test_reporter_report_newrelic_calls_newrelic_record_exception(self, newrelic_agent):
        reporter = Reporter()
        reporter._report_newrelic({
            'test_key1': 'test_value',
            'test_key2': 2
        })

        newrelic_agent.add_custom_attributes.assert_called_once_with([
            ('test_key1', 'test_value'),
            ('test_key2', 2)
        ])



import sys
import os, signal, traceback
import time
from contextlib import contextmanager
from pyworker.db import DBConnector
from pyworker.job import Job
from pyworker.logger import Logger
from pyworker.util import get_current_time, get_time_delta
from pyworker.reporter import Reporter

class TimeoutException(Exception): pass
class TerminatedException(Exception): pass

class Worker(object):
    def __init__(self, dbstring, logger=None,
                 extra_delayed_job_fields=None,
                 reported_attributes_prefix=''):
        super(Worker, self).__init__()
        self.logger = Logger(logger)
        self.logger.info('Starting pyworker...')
        self.database = DBConnector(dbstring, self.logger)
        self.sleep_delay = 10
        self.max_attempts = 3
        self.max_run_time = 3600
        self.queue_names = 'default'
        hostname = os.uname()[1]
        pid = os.getpid()
        self.name = 'host:%s pid:%d' % (hostname, pid)
        self.extra_delayed_job_fields = extra_delayed_job_fields

        # Configure application reporter if ENV variables set
        self.reporter = None
        NEW_RELIC_LICENSE_KEY = os.environ.get("NEW_RELIC_LICENSE_KEY")
        NEW_RELIC_APP_NAME = os.environ.get("NEW_RELIC_APP_NAME")

        # Register application reporter if configured
        if NEW_RELIC_LICENSE_KEY and NEW_RELIC_APP_NAME:
            self.reporter = Reporter(
                attribute_prefix=reported_attributes_prefix, logger=self.logger)

    @contextmanager
    def _time_limit(self, seconds):
        def signal_handler(signum, frame):
            raise TimeoutException(('Execution expired. Either do ' + \
                'the job faster or raise max_run_time > %d seconds') % \
                self.max_run_time)
        signal.signal(signal.SIGALRM, signal_handler)
        signal.alarm(seconds)
        try:
            yield
        finally:
            signal.alarm(0)

    @contextmanager
    def _terminatable(self):
        def signal_handler(signum, frame):
            signal_name = 'SIGTERM' if signum == 15 else 'SIGINT'
            self.logger.info('Received signal: %s' % signal_name)
            raise TerminatedException(signal_name)
        signal.signal(signal.SIGINT, signal_handler)
        signal.signal(signal.SIGTERM, signal_handler)
        yield

    @contextmanager
    def _instrument(self, job):

        def _latency(job_run_at):
            # we stick to get_current_time() to match the one used in the UPDATE query
            now = get_current_time()

            # Difference between when the job was scheduled `job_run_at`
            # and when the job actually started running `now`
            return (now - job_run_at).total_seconds()

        if self.reporter:
            latency = _latency(job.run_at)

            with self.reporter.recorder(job.job_name) as task:

                # Record custom attributes for the job transaction
                self.reporter.report(
                    job_id=job.job_id,
                    job_name=job.job_name,
                    job_queue=job.queue,
                    job_latency=latency,
                    job_attempts=job.attempts
                )

                # Record extra fields if configured
                self.logger.debug('job extra fields: %s' % job.extra_fields)
                if job.extra_fields is not None:
                    self.reporter.report(**job.extra_fields)

                yield task
        else:
            yield

    def run(self):
        # continuously check for new jobs on specified queue from db
        self._cursor = self.database.connect().cursor()
        with self._terminatable():
            while True:
                self.logger.debug('Picking up jobs...')
                job = self.get_job()
                self._current_job = job # used in signal handlers
                try:
                    if job is not None:
                        self.handle_job(job)
                    else: # sleep for a while before checking again for new jobs
                        time.sleep(self.sleep_delay)
                except TerminatedException:
                    break

            self.database.disconnect()

            # If configured shutdown reporter to upload data on shutdown
            if self.reporter:
                self.reporter.shutdown()

    def get_job(self):
        def get_job_row():
            now = get_current_time()
            expired = now - get_time_delta(seconds=self.max_run_time)
            now, expired = str(now), str(expired)
            queues = self.queue_names.split(',')
            queues = ', '.join(["'%s'" % q for q in queues])
            fields = ['id', 'attempts', 'run_at', 'queue', 'handler']
            if self.extra_delayed_job_fields:
                fields += self.extra_delayed_job_fields
            fields = ', '.join(fields)
            query = '''
            UPDATE delayed_jobs SET locked_at = '%s', locked_by = '%s'
            WHERE id IN (SELECT delayed_jobs.id FROM delayed_jobs
                WHERE ((run_at <= '%s'
                AND (locked_at IS NULL OR locked_at < '%s')
                OR locked_by = '%s') AND failed_at IS NULL)
                AND delayed_jobs.queue IN (%s)
            ORDER BY priority ASC, run_at ASC LIMIT 1 FOR UPDATE) RETURNING
                %s
            ''' % (now, self.name, now, expired, self.name, queues, fields)
            self.logger.debug('query: %s' % query)
            self._cursor.execute(query)
            return self._cursor.fetchone()

        job_row = get_job_row()
        if job_row:
            return Job.from_row(job_row, max_attempts=self.max_attempts,
                database=self.database, logger=self.logger,
                extra_fields=self.extra_delayed_job_fields,
                reporter=self.reporter)
        else:
            return None

    def handle_job(self, job):
        if job is None:
            return
        with self._instrument(job):
            start_time = time.time()
            error = failed = False
            caught_exc_info = None
            try:
                if job.abstract:
                    raise ValueError(('Unsupported Job: %s, please import it ' \
                        + 'before you can handle it') % job.class_name)
                else:
                    self.logger.info('Running Job %d' % job.job_id)
                    with self._time_limit(self.max_run_time):
                        job.before()
                        job.run()
                        job.after()
                    job.success()
                    job.remove()
            except Exception as exception:
                error = True
                caught_exc_info = sys.exc_info() # tuple of type, value, traceback
                # handle error
                error_str = traceback.format_exc()
                failed = job.set_error_unlock(error_str)
                # if that was a termination error, bubble up to caller
                if type(exception) == TerminatedException:
                    raise exception
            finally:
                # report error status
                if self.reporter:
                    self.reporter.report_raw(error=error)
                    self.reporter.report(job_failure=failed)
                    if caught_exc_info:
                        self.reporter.record_exception(caught_exc_info)
                time_diff = time.time() - start_time
                self.logger.info('Job %d finished in %d seconds' % \
                    (job.job_id, time_diff))



import json
from contextlib import contextmanager
import newrelic.agent


class Reporter(object):

    def __init__(self, attribute_prefix='', logger=None):
        self._prefix = attribute_prefix
        self._logger = logger
        if self._logger:
            self._logger.info('Reporter: initializing NewRelic')
        newrelic.agent.initialize()
        self._newrelic_app = newrelic.agent.register_application()

    def report(self, **attributes):
        # flatten attributes
        attributes = self._flatten_attributes(attributes)
        # format attributes
        attributes = self._format_attributes(attributes)
        self.report_raw(**attributes)

    def report_raw(self, **attributes):
        # report to NewRelic
        self._report_newrelic(attributes)

    @contextmanager
    def recorder(self, name):
        with newrelic.agent.BackgroundTask(
                application=self._newrelic_app,
                name=name,
                group='DelayedJob') as task:
            yield task

    def shutdown(self):
        newrelic.agent.shutdown_agent()

    def record_exception(self, exc_info):
        newrelic.agent.notice_error(error=exc_info)

    def _flatten_attributes(self, attributes):
        # flatten nested dict attributes
        flattened_attributes = {}
        for key, value in attributes.items():
            if type(value) == dict:
                for nested_key, nested_value in value.items():
                    flattened_attributes[nested_key] = nested_value
            else:
                flattened_attributes[key] = value
        return flattened_attributes

    def _format_attributes(self, attributes):
        # prefix then convert all attribute keys to camelCase
        # ensure values types are supported or json dump them
        return {
            self._prefix + self._to_camel_case(key): self._convert_value(value)
            for key, value in attributes.items()
            if key is not None and value is not None
        }

    @staticmethod
    def _to_camel_case(string):
        return string[0]+string.title()[1:].replace("-","").replace("_","").replace(" ","")

    @staticmethod
    def _convert_value(value):
        if type(value) not in [str, int, float, bool]:
            return json.dumps(value)
        return value

    def _report_newrelic(self, attributes):
        if self._logger:
            self._logger.debug('Reporter: reporting to NewRelic: %s' % attributes)
        # convert attributes dict to list of tuples
        attributes = list(attributes.items())
        newrelic.agent.add_custom_attributes(attributes)



import datetime
from unittest import TestCase
from unittest.mock import patch, MagicMock
from pyworker.job import Job, get_current_time, get_time_delta


class RegisteredJob(Job): # matching the registered class fixture
    def run(self):
        pass


class TestJob(TestCase):
    def setUp(self):
        self.mock_job_id = 1
        self.mock_attempts = 0
        self.mock_run_at = datetime.datetime(2023, 10, 7, 0, 0, 1)
        self.mock_queue = 'default'
        self.mock_max_attempts = 5
        self.mock_now = datetime.datetime(2023, 10, 7, 0, 0, 0)
        self.mock_extra_fields = {
            'extra_field1': 'extra_field1_value',
            'extra_field2': 100,
            'extra_field3': True,
            'extra_field4': {'a': [1, 2, 3]},
            'extra_field5': None
        }

    def tearDown(self):
        pass

    def load_fixture(self, filename):
        with open('tests/fixtures/%s' % filename) as f:
            return f.read()

    def load_job(self, filename):
        mock_handler = self.load_fixture(filename)
        mock_row = (
            self.mock_job_id,
            self.mock_attempts,
            self.mock_run_at,
            self.mock_queue,
            mock_handler
        )
        return Job.from_row(mock_row,
                            self.mock_max_attempts,
                            MagicMock(), MagicMock())

    def load_job_with_extra_fields(self, filename):
        mock_handler = self.load_fixture(filename)
        mock_row = (
            self.mock_job_id,
            self.mock_attempts,
            self.mock_run_at,
            self.mock_queue,
            mock_handler,
            *self.mock_extra_fields.values()
        )
        return Job.from_row(mock_row,
                            self.mock_max_attempts,
                            MagicMock(), MagicMock(),
                            extra_fields=self.mock_extra_fields.keys())

    def load_unregistered_job(self):
        return self.load_job('handler_unregistered.yaml')

    def load_unregistered_job_with_extra_fields(self):
        return self.load_job_with_extra_fields('handler_unregistered.yaml')

    def load_unregistered_job_with_reporter(self, reporter):
        job = self.load_unregistered_job()
        job.reporter = reporter
        return job

    def load_registered_job(self):
        job = self.load_job('handler_registered.yaml')
        job.error = MagicMock()
        job.failure = MagicMock()
        job._update_job = MagicMock()
        return job

    def load_registered_job_with_extra_fields(self):
        return self.load_job_with_extra_fields('handler_registered.yaml')

    def load_registered_job_with_reporter(self, reporter):
        job = self.load_registered_job()
        job.reporter = reporter
        return job

    def load_registered_job_with_attempts_exceeded(self):
        job = self.load_registered_job()
        job.attempts = self.mock_max_attempts - 1
        return job

    #********** .from_row tests **********#

    def test_from_row_when_unregistered_class_returns_abstract_job_instance(self):
        job = self.load_unregistered_job()

        self.assertEqual(job.class_name, 'UnregisteredJob')
        self.assertEqual(job.abstract, True)

    def test_from_row_when_unregistered_class_returns_job_instance_without_attributes(self):
        job = self.load_unregistered_job()

        self.assertEqual(job.job_id, self.mock_job_id)
        self.assertEqual(job.attempts, self.mock_attempts)
        self.assertEqual(job.run_at, self.mock_run_at)
        self.assertEqual(job.queue, self.mock_queue)
        self.assertEqual(job.max_attempts, self.mock_max_attempts)
        self.assertIsNone(job.extra_fields)
        self.assertIsNone(job.attributes)
        self.assertIsNone(job.reporter)

    def test_from_row_when_unregistered_class_returns_job_instance_with_extra_fields(self):
        job = self.load_unregistered_job_with_extra_fields()

        self.assertDictEqual(job.extra_fields, self.mock_extra_fields)

    def test_from_row_when_unregistered_class_returns_abstract_job_instance_with_reporter(self):
        mock_reporter = MagicMock()
        job = self.load_unregistered_job_with_reporter(mock_reporter)

        self.assertEqual(job.reporter, mock_reporter)

    def test_from_row_when_registered_class_returns_concrete_job_instance(self):
        job = self.load_registered_job()

        self.assertEqual(job.class_name, 'RegisteredJob')
        self.assertEqual(job.abstract, False)

    def test_from_row_when_registered_class_returns_job_instance_with_attributes(self):
        job = self.load_registered_job()

        self.assertEqual(job.job_id, self.mock_job_id)
        self.assertEqual(job.attempts, self.mock_attempts)
        self.assertEqual(job.run_at, self.mock_run_at)
        self.assertEqual(job.queue, self.mock_queue)
        self.assertEqual(job.max_attempts, self.mock_max_attempts)
        self.assertIsNone(job.extra_fields)
        # below attributes match the registered class fixture
        self.assertDictEqual(job.attributes, {
            'id': 100,
            'title': 'review title',
            'description': 'review description\nmultiline\n',
            'total_articles': 1000,
            'is_blind': True
        })
        self.assertIsNone(job.reporter)

    def test_from_row_when_registered_class_returns_job_instance_with_extra_fields(self):
        job = self.load_registered_job_with_extra_fields()

        self.assertDictEqual(job.extra_fields, self.mock_extra_fields)

    def test_from_row_when_registered_class_returns_concrete_job_instance_with_reporter(self):
        mock_reporter = MagicMock()
        job = self.load_registered_job_with_reporter(mock_reporter)

        self.assertEqual(job.reporter, mock_reporter)

    #********** .set_error_unlock tests **********#

    def assert_job_updated_field(self, job, field, value):
        job._update_job.assert_called_once()
        setters = job._update_job.call_args[0][0]
        values = job._update_job.call_args[0][1]
        index = setters.index("%s = %%s" % field)
        assert index >= 0
        assert values[index] == value

    def assert_job_non_updated_field(self, job, field):
        job._update_job.assert_called_once()
        setters = job._update_job.call_args[0][0]
        assert "%s = %%s" % field not in setters

    def assert_job_updated_run_at(self, job, attempts, expected_value):
        job.attempts = attempts
        expected_value = expected_value.strftime('%Y-%m-%d %H:%M:%S')

        job.set_error_unlock('some error')

        self.assert_job_updated_field(job, 'run_at', expected_value)

    # max attempts not exceeded

    ## error and failure hooks
    def test_set_error_unlock_if_max_attempts_not_exceeded_calls_error_hook_only(self):
        job = self.load_registered_job()

        job.set_error_unlock('some error')

        job.error.assert_called_once_with('some error')
        job.failure.assert_not_called()

    ## attempts
    def test_set_error_unlock_if_max_attempts_not_exceeded_increments_attempts(self):
        job = self.load_registered_job()
        expected_value = job.attempts + 1

        job.set_error_unlock('some error')

        self.assert_job_updated_field(job, 'attempts', expected_value)

    ## run_at
    @patch('pyworker.job.get_current_time')
    def test_set_error_unlock_if_max_attempts_not_exceeded_updates_run_at_exponentially_when_attempts_0(
            self, mock_get_current_time):
        mock_get_current_time.return_value = self.mock_now
        job = self.load_registered_job()

        self.assert_job_updated_run_at(job, attempts=0, expected_value=datetime.datetime(2023, 10, 7, 0, 0, 6))

    @patch('pyworker.job.get_current_time')
    def test_set_error_unlock_if_max_attempts_not_exceeded_updates_run_at_exponentially_when_attempts_1(
            self, mock_get_current_time):
        mock_get_current_time.return_value = self.mock_now
        job = self.load_registered_job()

        self.assert_job_updated_run_at(job, attempts=1, expected_value=datetime.datetime(2023, 10, 7, 0, 0, 21))

    @patch('pyworker.job.get_current_time')
    def test_set_error_unlock_if_max_attempts_not_exceeded_updates_run_at_exponentially_when_attempts_2(
            self, mock_get_current_time):
        mock_get_current_time.return_value = self.mock_now
        job = self.load_registered_job()

        self.assert_job_updated_run_at(job, attempts=2, expected_value=datetime.datetime(2023, 10, 7, 0, 1, 26))

    @patch('pyworker.job.get_current_time')
    def test_set_error_unlock_if_max_attempts_not_exceeded_updates_run_at_exponentially_when_attempts_3(
            self, mock_get_current_time):
        mock_get_current_time.return_value = self.mock_now
        job = self.load_registered_job()

        self.assert_job_updated_run_at(job, attempts=3, expected_value=datetime.datetime(2023, 10, 7, 0, 4, 21))

    ## failed_at
    @patch('pyworker.job.get_current_time')
    def test_set_error_unlock_if_max_attempts_not_exceeded_does_not_update_failed_at(
            self, mock_get_current_time):
        mock_get_current_time.return_value = self.mock_now
        job = self.load_registered_job()

        job.set_error_unlock('some error')

        self.assert_job_non_updated_field(job, 'failed_at')

    ## locked_at
    def test_set_error_unlock_if_max_attempts_not_exceeded_nullifies_job_locked_at(self):
        job = self.load_registered_job()

        job.set_error_unlock('some error')

        self.assert_job_updated_field(job, 'locked_at', None)

    ## locked_by
    def test_set_error_unlock_if_max_attempts_not_exceeded_nullifies_job_locked_by(self):
        job = self.load_registered_job()

        job.set_error_unlock('some error')

        self.assert_job_updated_field(job, 'locked_by', None)

    ## last_error
    def test_set_error_unlock_if_max_attempts_not_exceeded_updates_last_error(self):
        job = self.load_registered_job()

        job.set_error_unlock('some error')

        self.assert_job_updated_field(job, 'last_error', 'some error')

    ## returns
    def test_set_error_unlock_if_max_attempts_not_exceeded_returns_false(self):
        job = self.load_registered_job()

        self.assertFalse(job.set_error_unlock('some error'))

    ## max attempts exceeded

    ## error and failure hooks
    def test_set_error_unlock_if_max_attempts_exceeded_calls_error_and_failure_hooks(self):
        job = self.load_registered_job_with_attempts_exceeded()

        job.set_error_unlock('some error')

        job.error.assert_called_once_with('some error')
        job.failure.assert_called_once_with('some error')

    ## attempts
    def test_set_error_unlock_if_max_attempts_exceeded_increments_attempts(self):
        job = self.load_registered_job_with_attempts_exceeded()
        expected_value = job.attempts + 1

        job.set_error_unlock('some error')

        self.assert_job_updated_field(job, 'attempts', expected_value)

    ## run_at
    def test_set_error_unlock_if_max_attempts_exceeded_does_not_update_run_at(self):
        job = self.load_registered_job_with_attempts_exceeded()

        job.set_error_unlock('some error')

        self.assert_job_non_updated_field(job, 'run_at')

    ## failed_at
    @patch('pyworker.job.get_current_time')
    def test_set_error_unlock_if_max_attempts_exceeded_updates_failed_at(
            self, mock_get_current_time):
        mock_get_current_time.return_value = self.mock_now
        job = self.load_registered_job_with_attempts_exceeded()

        job.set_error_unlock('some error')

        self.assert_job_updated_field(job, 'failed_at', mock_get_current_time.return_value)

    ## locked_at
    def test_set_error_unlock_if_max_attempts_exceeded_nullifies_job_locked_at(self):
        job = self.load_registered_job_with_attempts_exceeded()

        job.set_error_unlock('some error')

        self.assert_job_updated_field(job, 'locked_at', None)

    ## locked_by
    def test_set_error_unlock_if_max_attempts_exceeded_nullifies_job_locked_by(self):
        job = self.load_registered_job_with_attempts_exceeded()

        job.set_error_unlock('some error')

        self.assert_job_updated_field(job, 'locked_by', None)

    ## last_error
    def test_set_error_unlock_if_max_attempts_exceeded_updates_last_error(self):
        job = self.load_registered_job_with_attempts_exceeded()

        job.set_error_unlock('some error')

        self.assert_job_updated_field(job, 'last_error', 'some error')

    ## returns
    def test_set_error_unlock_if_max_attempts_exceeded_returns_true(self):
        job = self.load_registered_job_with_attempts_exceeded()

        self.assertTrue(job.set_error_unlock('some error'))




import datetime
from unittest import TestCase
from unittest.mock import patch, MagicMock
from pyworker.worker import Worker, TerminatedException

class TestWorker(TestCase):
    @patch('pyworker.worker.DBConnector')
    def setUp(self, mock_db):
        self.worker = Worker('dummy')
        self.mock_db = mock_db
        mock_db.connect = MagicMock()
        mock_db.disconnect = MagicMock()
        mocked_run_at = datetime.datetime(2023, 10, 7, 0, 0, 1)
        self.mocked_now = datetime.datetime(2023, 10, 7, 0, 0, 10)
        self.mocked_latency = 9 # seconds: mocked_now - mocked_run_at
        self.mock_job = MagicMock(
            abstract=False,
            job_id=1,
            job_name='test_job',
            queue='default',
            attempts=0,
            run_at=mocked_run_at)
        self.mock_extra_fields = {
            'extra_field1_str': 'extra_field1_value',
            'extra_field2_int': 100,
            'extra_field3_float': 1.1,
            'extra_field4_bool': True,
            'extra_field5_bool': False,
            'extra_field6_json': {'a': [1, 2, 3]},
            'extra_field7_none': None
        }

    def tearDown(self):
        pass

    #********** __init__ tests **********#

    @patch('pyworker.worker.os.uname', return_value=['pytest', 'localhost'])
    @patch('pyworker.worker.os.getpid', return_value=1234)
    @patch('pyworker.worker.DBConnector')
    def test_worker_init(self, mock_db, *_):
        worker = Worker('dummy')

        self.assertEqual(worker.database, mock_db.return_value)
        self.assertEqual(worker.sleep_delay, 10)
        self.assertEqual(worker.max_attempts, 3)
        self.assertEqual(worker.max_run_time, 3600)
        self.assertEqual(worker.queue_names, 'default')
        self.assertEqual(worker.name, 'host:localhost pid:1234')
        self.assertIsNone(worker.extra_delayed_job_fields)
        self.assertIsNone(worker.reporter)

    @patch('pyworker.worker.DBConnector')
    def test_worker_init_with_extra_delayed_job_fields(self, *_):
        worker = Worker('dummy', extra_delayed_job_fields=self.mock_extra_fields.keys())

        self.assertEqual(worker.extra_delayed_job_fields, self.mock_extra_fields.keys())

    @patch('pyworker.worker.DBConnector')
    @patch('pyworker.worker.os.environ', {
        'NEW_RELIC_LICENSE_KEY': 'test', 'NEW_RELIC_APP_NAME': 'test'})
    @patch('pyworker.worker.Reporter')
    def test_worker_init_with_reporter(self, mock_reporter, *_):
        mock_reporter.return_value = MagicMock()
        worker = Worker('dummy', reported_attributes_prefix='test_prefix')

        self.assertEqual(worker.reporter, mock_reporter.return_value)
        mock_reporter.assert_called_once_with(
            attribute_prefix='test_prefix', logger=worker.logger)

    #********** .run tests **********#

    @patch('pyworker.worker.Worker.get_job', return_value=None)
    # make sleep raise an exception to stop the loop
    @patch('pyworker.worker.time.sleep', side_effect=TerminatedException('SIGTERM'))
    def test_worker_run_connects_to_and_disconnects_from_database(self, *_):
        self.worker.run()

        self.worker.database.connect.assert_called_once_with()
        self.worker.database.disconnect.assert_called_once_with()

    @patch('pyworker.worker.Worker.get_job', return_value=None)
    @patch('pyworker.worker.time.sleep', side_effect=TerminatedException('SIGTERM'))
    def test_worker_run_shuts_down_reporter(self, *_):
        self.worker.reporter = MagicMock()

        self.worker.run()

        self.worker.reporter.shutdown.assert_called_once_with()

    @patch('pyworker.worker.time.sleep', side_effect=TerminatedException('SIGTERM'))
    @patch('pyworker.worker.Worker.get_job', return_value=None)
    def test_worker_run_when_no_jobs_found_sleeps(self, mock_get_job, mock_time_sleep):
        self.worker.run()

        mock_get_job.assert_called_once_with()
        mock_time_sleep.assert_called_once_with(self.worker.sleep_delay)

    @patch('pyworker.worker.Worker.handle_job', side_effect=TerminatedException('SIGTERM'))
    @patch('pyworker.worker.Worker.get_job', return_value=MagicMock())
    def test_worker_run_when_job_found_handles_job(self, mock_get_job, mock_handle_job):
        self.worker.run()

        mock_get_job.assert_called_once_with()
        mock_handle_job.assert_called_once_with(mock_get_job.return_value)

    #********** .handle_job tests **********#

    def assert_instrument_context_reports_custom_attributes(self, job, reporter):
        reporter.recorder.assert_called_once()
        reporter.report.assert_any_call(
            job_id=job.job_id,
            job_name=job.job_name,
            job_queue=job.queue,
            job_latency=self.mocked_latency,
            job_attempts=job.attempts
        )
        if job.extra_fields is not None:
            reporter.report.assert_any_call(**job.extra_fields)

    def test_worker_handle_job_when_job_is_none_does_nothing(self):
        self.worker.handle_job(None) # no error raised

    def test_worker_handle_job_when_job_is_unsupported_type_sets_error(self):
        job = self.mock_job
        job.abstract = True

        self.worker.handle_job(job)

        job.set_error_unlock.assert_called_once()
        assert 'Unsupported Job' in job.set_error_unlock.call_args[0][0]

    @patch('pyworker.worker.get_current_time')
    def test_worker_handle_job_when_job_is_unsupported_type_reports_error(
            self, get_current_time):
        get_current_time.return_value = self.mocked_now
        job = self.mock_job
        job.abstract = True
        job.set_error_unlock.return_value = False
        reporter = MagicMock()
        self.worker.reporter = reporter

        self.worker.handle_job(job)

        self.assert_instrument_context_reports_custom_attributes(job, reporter)
        reporter.record_exception.assert_called_once()
        reporter.report.assert_any_call(job_failure=False)
        reporter.report_raw.assert_any_call(error=True)

    @patch('pyworker.worker.get_current_time')
    def test_worker_handle_job_when_job_is_unsupported_type_reports_extra_fields(
            self, get_current_time):
        get_current_time.return_value = self.mocked_now
        job = self.mock_job
        job.abstract = True
        job.extra_fields = self.mock_extra_fields
        reporter = MagicMock()
        self.worker.reporter = reporter

        self.worker.handle_job(job)

        self.assert_instrument_context_reports_custom_attributes(job, reporter)

    def test_worker_handle_job_calls_all_hooks_then_removes_from_queue(self):
        self.worker.handle_job(self.mock_job)

        self.mock_job.before.assert_called_once()
        self.mock_job.run.assert_called_once()
        self.mock_job.after.assert_called_once()
        self.mock_job.success.assert_called_once()

        self.mock_job.remove.assert_called_once()

    @patch('pyworker.worker.get_current_time')
    def test_worker_handle_job_when_no_errors_reports_success(
            self, get_current_time):
        get_current_time.return_value = self.mocked_now
        job = self.mock_job
        job.extra_fields = self.mock_extra_fields
        reporter = MagicMock()
        self.worker.reporter = reporter

        self.worker.handle_job(job)

        reporter.record_exception.assert_not_called()
        reporter.report.assert_any_call(job_failure=False)
        reporter.report_raw.assert_any_call(error=False)
        self.assert_instrument_context_reports_custom_attributes(job, reporter)

    def test_worker_handle_job_when_error_sets_error_and_unlocks_job(self):
        job = self.mock_job
        job.run.side_effect = Exception('test error')

        self.worker.handle_job(job)

        job.set_error_unlock.assert_called_once()
        assert 'test error' in job.set_error_unlock.call_args[0][0]
        job.remove.assert_not_called()

    @patch('pyworker.worker.get_current_time')
    def test_worker_handle_job_when_error_reports_error(self,
            get_current_time):
        get_current_time.return_value = self.mocked_now
        job = self.mock_job
        job.set_error_unlock.return_value = False
        job.run.side_effect = Exception('test error')
        job.extra_fields = self.mock_extra_fields
        reporter = MagicMock()
        self.worker.reporter = reporter

        self.worker.handle_job(job)

        self.assert_instrument_context_reports_custom_attributes(job, reporter)
        reporter.record_exception.assert_called_once()
        reporter.report.assert_any_call(job_failure=False)
        reporter.report_raw.assert_any_call(error=True)
        job.remove.assert_not_called()

    @patch('pyworker.worker.get_current_time')
    def test_worker_handle_job_when_permanent_error_reports_failure(
            self, get_current_time):
        get_current_time.return_value = self.mocked_now
        job = self.mock_job
        job.set_error_unlock.return_value = True
        job.run.side_effect = Exception('test error')
        job.extra_fields = self.mock_extra_fields
        reporter = MagicMock()
        self.worker.reporter = reporter

        self.worker.handle_job(job)

        self.assert_instrument_context_reports_custom_attributes(job, reporter)
        reporter.record_exception.assert_called_once()
        reporter.report.assert_any_call(job_failure=True)
        reporter.report_raw.assert_any_call(error=True)
        job.remove.assert_not_called()

    def test_worker_handle_job_when_error_is_termination_error_bubbles_up(self):
        job = self.mock_job
        job.run.side_effect = TerminatedException('SIGTERM')

        with self.assertRaises(TerminatedException):
            self.worker.handle_job(job)



from setuptools import setup

requirements = [
    'psycopg2-binary>=2',
    'python-dateutil>=2',
    'PyYAML>=3',
    'newrelic>=8.3.0'
]

setup(
    name = "rubydj-pyworker",
    version = '1.1.0',
    description="A pure Python worker for Ruby-based DelayedJobs with Newrelic reporting.",
    author="Rayyan Systems, Inc.",
    author_email="support@rayyan.ai",
    url="https://github.com/rayyansys/pyworker",
    license="MIT",
    packages=['pyworker'],
    install_requires = requirements
)
